/*
GNU General Public License

Grafito - Simple Grafitti Board App
Copyright (C) 2000  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.grafito;

import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Checkbox;
import java.awt.CheckboxGroup;
import java.awt.Choice;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.Label;
import java.awt.MediaTracker;
import java.awt.Panel;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Hashtable;

import com.sun.image.codec.jpeg.JPEGCodec;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

/** Grafito
  * App for providing a simple grafitti board
  *
  * @author Howard Kistler
  * @version 0.2
  *
  * REQUIREMENTS
  * JDK 1.2.2 or higher
  *
  * VERSION HISTORY
  * 0.1  (09/14/2000) - initial creation (09/14/2000)
  * 0.2  (11/03/2000) - added GrafitoCanvas to better control painting
  *                   - added background image chooser
  *                   - implemented MediaTracker to make sure images preload (more important than it seems)
  */

public class Grafito extends Frame implements ActionListener, WindowListener, MouseListener, MouseMotionListener, ItemListener
{
	// Drawing canvas
	private GrafitoCanvas canvas = new GrafitoCanvas();

	// Image to store drawing
	private Image imageStore;

	// Optional background image
	private Image imageBack;
	private int currImage = 0;

	// Default app size variables
	protected int appXDimension  = 375;
	protected int appYDimension  = 300;
	protected int menuXDimension = 75;
	protected int cnvsXDimension = appXDimension - menuXDimension;

	// Current paint setting variables
	private Color currColor = Color.black;
	private Color backColor = Color.white;

	// UI components
	private Checkbox     chkBrushSmall;
	private Checkbox     chkBrushMedium;
	private Checkbox     chkBrushLarge;
	private Choice       chsPaintColor;
	private Choice       chsBackground;
	private MediaTracker mt;

	// Collections
	private Color[]   forecolors;
	private Image[]   backgrounds;
	private Hashtable htParams;

	/* Constructor */
	public Grafito(String paramfile)
	{
		super("Grafito");

		mt = new MediaTracker(this);

		int canvasWidth  = -1;
		int canvasHeight = -1;

		htParams = new Hashtable();
		if(paramfile != null)
		{
			loadParams(paramfile);
		}

		// Fetch parameters
		String appXparam  = getParameter("APPWIDTH");
		String appYparam  = getParameter("APPHEIGHT");
		String menuXparam = getParameter("MENUWIDTH");
		String cnvsXparam = getParameter("CANVASWIDTH");
		String cnvsYparam = getParameter("CANVASHEIGHT");
		String imgNames   = getParameter("IMAGENAMES");
		String imgFiles   = getParameter("IMAGEFILES");
		String imgBack    = getParameter("IMAGEBACK");
		String bgColor    = getParameter("COLORBACK");
		String clrNames   = getParameter("COLORNAMES");
		String clrValues  = getParameter("COLORVALUES");

		// Process parameters
		if(appXparam != null)
		{
			try
			{
				appXDimension = Integer.parseInt(appXparam);
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(appYparam != null)
		{
			try
			{
				appYDimension = Integer.parseInt(appYparam);
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(menuXparam != null)
		{
			try
			{
				menuXDimension = Integer.parseInt(menuXparam);
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(cnvsXparam != null)
		{
			try
			{
				canvasWidth = Integer.parseInt(cnvsXparam);
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(cnvsXparam != null)
		{
			try
			{
				canvasHeight = Integer.parseInt(cnvsXparam);
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(canvasWidth > -1)
		{
			cnvsXDimension = canvasWidth;
			appXDimension  = cnvsXDimension + menuXDimension;
		}
		if(canvasHeight > -1)
		{
			appYDimension = canvasHeight;
		}
		if((imgNames != null) && (imgFiles != null))
		{
			chsBackground = new Choice();
			chsBackground.addItemListener(this);
			chsBackground.add("None");

			// Parse image names
			while(imgNames.indexOf("|") > -1)
			{
				String iName = imgNames.substring(0, imgNames.indexOf("|"));
				chsBackground.add(iName);
				imgNames = imgNames.substring(imgNames.indexOf("|") + 1, imgNames.length());
			}
			if(imgNames.length() > 0)
			{
				chsBackground.add(imgNames);
			}

			// Load image files
			backgrounds = new Image[chsBackground.getItemCount()];
			int imagecount = 0;
			backgrounds[imagecount] = (Image)null;
			imagecount++;
			while((imgFiles.indexOf("|") > -1) && (imagecount < chsBackground.getItemCount()))
			{
				String iFile = imgFiles.substring(0, imgFiles.indexOf("|"));
				backgrounds[imagecount] = loadImage(iFile);
				imagecount++;
				imgFiles = imgFiles.substring(imgFiles.indexOf("|") + 1, imgFiles.length());
			}
			if((imgFiles.length() > 0) && (imagecount < chsBackground.getItemCount()))
			{
				backgrounds[imagecount] = loadImage(imgFiles);
				imagecount++;
			}
		}
		if(bgColor != null)
		{
			backColor = parseColor(bgColor);
		}
		if(imgBack != null)
		{
			MediaTracker mt = new MediaTracker(this);
			Image tempImage = loadImage(imgBack);
			mt.addImage(tempImage, 0);
			try
			{
				mt.waitForID(0);
			}
			catch(Exception e)
			{
			}
			imageBack = tempImage;
		}

		/* Set up the controls */
		Button btnFileSave    = new Button("Save");    btnFileSave.setActionCommand("save");       btnFileSave.addActionListener(this);
		Button btnFileRefresh = new Button("Refresh"); btnFileRefresh.setActionCommand("refresh"); btnFileRefresh.addActionListener(this);
		Button btnFileClear   = new Button("Clear");   btnFileClear.setActionCommand("clear");     btnFileClear.addActionListener(this);

		Label lblBrush = new Label("BRUSH", Label.CENTER);

		CheckboxGroup cgrpBrushes = new CheckboxGroup();
		chkBrushSmall  = new Checkbox("Small",  cgrpBrushes, false);
		chkBrushMedium = new Checkbox("Medium", cgrpBrushes, true);
		chkBrushLarge  = new Checkbox("Large",  cgrpBrushes, false);

		Label lblColor = new Label("COLOR", Label.CENTER);

		if((clrNames != null) && (clrValues != null))
		{
			chsPaintColor = new Choice();
			chsPaintColor.addItemListener(this);

			// Parse color names
			while(clrNames.indexOf("|") > -1)
			{
				String cName = clrNames.substring(0, clrNames.indexOf("|"));
				chsPaintColor.add(cName);
				clrNames = clrNames.substring(clrNames.indexOf("|") + 1, clrNames.length());
			}
			if(clrNames.length() > 0)
			{
				chsPaintColor.add(clrNames);
			}

			// Parse color values
			forecolors = new Color[chsPaintColor.getItemCount()];
			int colorcount = 0;
			while((clrValues.indexOf("|") > -1) && (colorcount < chsPaintColor.getItemCount()))
			{
				String cValue = clrValues.substring(0, clrValues.indexOf("|"));
				forecolors[colorcount] = parseColor(cValue);
				colorcount++;
				clrValues = clrValues.substring(clrValues.indexOf("|") + 1, clrValues.length());
			}
			if((clrValues.length() > 0) && (colorcount < chsPaintColor.getItemCount()))
			{
				forecolors[colorcount] = parseColor(clrValues);
				colorcount++;
			}

			chsPaintColor.select(0);
			currColor = forecolors[0];
		}

		Panel pnlButtons = new Panel();
		pnlButtons.setBackground(new Color(232, 232, 232));
		int rowcount = 7 + (chsPaintColor == null ? 0 : 2) + (chsBackground == null ? 0 : 2);
		pnlButtons.setLayout(new GridLayout(rowcount, 1));
		if(chsBackground != null)
		{
			Label lblBack = new Label("IMAGE", Label.CENTER);
			pnlButtons.add(lblBack);
			pnlButtons.add(chsBackground);
		}
		pnlButtons.add(lblBrush);
		pnlButtons.add(chkBrushSmall);
		pnlButtons.add(chkBrushMedium);
		pnlButtons.add(chkBrushLarge);
		if(chsPaintColor != null)
		{
			pnlButtons.add(lblColor);
			pnlButtons.add(chsPaintColor);
		}
		pnlButtons.add(btnFileClear);
		pnlButtons.add(btnFileRefresh);
		pnlButtons.add(btnFileSave);
		pnlButtons.setSize(menuXDimension, appYDimension);
		pnlButtons.setBounds(0, 0, menuXDimension, appYDimension);

		canvas.addMouseListener(this);
		canvas.addMouseMotionListener(this);
		canvas.setSize(cnvsXDimension, appYDimension);
		canvas.setBounds(menuXDimension, 0, cnvsXDimension, appYDimension);
		canvas.setBackground(backColor);
		canvas.updateColor(backColor);
		canvas.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));

		this.addWindowListener(this);
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.setLayout(new BorderLayout());
		this.add(canvas, BorderLayout.CENTER);
		this.add(pnlButtons, BorderLayout.WEST);
		this.setSize(appXDimension, appYDimension);
		this.show();
	}

	/* ActionListener method */
	public void actionPerformed(ActionEvent ae)
	{
		if(ae.getActionCommand().equals("exit"))
		{
			this.dispose();
			System.exit(0);
		}
		else if(ae.getActionCommand().equals("save"))
		{
			writeToJPEG("grafito.jpg");
			writeToGIF("grafitoapp.gif");
		}
		else if(ae.getActionCommand().equals("clear"))
		{
			clearImage();
			paint(this.getGraphics());
		}
		else if(ae.getActionCommand().equals("refresh"))
		{
			paint(this.getGraphics());
		}
	}

	/* WindowListener methods */
	public void windowClosing(WindowEvent we)
	{
		this.dispose();
		System.exit(0);
	}
	public void windowOpened(WindowEvent we)      { ; }
	public void windowClosed(WindowEvent we)      { ; }
	public void windowActivated(WindowEvent we)   { ; }
	public void windowDeactivated(WindowEvent we) { ; }
	public void windowIconified(WindowEvent we)   { ; }
	public void windowDeiconified(WindowEvent we) { ; }

	/* MouseListener methods */
	public void mousePressed(MouseEvent me)  { paintDot(me); }
	public void mouseClicked(MouseEvent me)  { ; }
	public void mouseEntered(MouseEvent me)  { ; }
	public void mouseExited(MouseEvent me)   { ; }
	public void mouseReleased(MouseEvent me) { ; }

	/* MouseMotionListener methods */
	public void mouseDragged(MouseEvent me) { paintDot(me); }
	public void mouseMoved(MouseEvent me)   { ; }

	/* ItemListener method */
	public void itemStateChanged(ItemEvent e)
	{
		if(e.getSource() == chsBackground)
		{
			int i = chsBackground.getSelectedIndex();
			if(i != currImage)
			{
				imageBack = backgrounds[i];
	
				clearImage();
			}
			currImage = i;
		}
		else if(e.getSource() == chsPaintColor)
		{
			currColor = forecolors[chsPaintColor.getSelectedIndex()];
		}
	}

	/* Initialise the image */
	public void initForPaint()
	{
		imageStore = (Image)createImage(cnvsXDimension, appYDimension); 
		clearImage();
	}

	/* Clear the image */
	public void clearImage()
	{
		Graphics gimg = imageStore.getGraphics();
		gimg.setColor(backColor);
		gimg.fillRect(0, 0, cnvsXDimension, appYDimension);
		if(imageBack != null)
		{
			gimg.drawImage(imageBack, 0, 0, this);
		}
		canvas.updateImage(imageStore);
		canvas.clearCanvas();
		paint(this.getGraphics());
	}

	/* Draw on the canvas */
	private void paintDot(MouseEvent me)
	{
		if(imageStore == null) { initForPaint(); }
		Graphics gimg = imageStore.getGraphics();
		gimg.setColor(currColor);

		if(chkBrushSmall.getState())
		{
			int minX = me.getX();
			int minY = me.getY();
			if(minX < 0) { minX = 0; }
			if(minY < 0) { minY = 0; }
			if(minX > cnvsXDimension) { minX = cnvsXDimension; }
			if(minY > appYDimension)  { minY = appYDimension; }
			gimg.drawLine(me.getX(), me.getY(), me.getX(), me.getY());
		}
		else if(chkBrushMedium.getState())
		{
			int minX = me.getX() - 2;
			int minY = me.getY() - 2;
			int maxX = 5;
			int maxY = 5;
			if(minX < 0) { minX = 0; }
			if(minY < 0) { minY = 0; }
			if(me.getX() + maxX > cnvsXDimension) { maxX = cnvsXDimension; }
			if(me.getY() + maxY > appYDimension)  { maxY = appYDimension; }
			gimg.fillRect(minX, minY, maxX, maxY);
		}
		else if(chkBrushLarge.getState())
		{
			int minX = me.getX() - 4;
			int minY = me.getY() - 4;
			int maxX = 9;
			int maxY = 9;
			if(minX < 0) { minX = 0; }
			if(minY < 0) { minY = 0; }
			if(me.getX() + maxX > cnvsXDimension) { maxX = cnvsXDimension; }
			if(me.getY() + maxY > appYDimension)  { maxY = appYDimension; }
			gimg.fillRect(minX, minY, maxX, maxY);
		}
		canvas.updateImage(imageStore);
	}

	/* JPEG writer */
	public void writeToJPEG(String filename)
	{
		try
		{ 
			FileOutputStream fos = new FileOutputStream(filename);
			JPEGImageEncoder jpegEncoder = JPEGCodec.createJPEGEncoder(fos);
			jpegEncoder.encode((java.awt.image.BufferedImage)imageStore);
			fos.flush();
			fos.close();
		} 
		catch(FileNotFoundException fnfe)
		{
			System.out.println(fnfe);
		}
		catch(IOException ioe)
		{
			System.out.println(ioe);
		}
	}

	/* GIF writer */
	public void writeToGIF(String filename)
	{
		try
		{
			java.awt.Dimension dim = new java.awt.Dimension(imageStore.getWidth(this), imageStore.getHeight(this));
			System.out.println("IMAGE SIZE : (" + dim.width + "x" + dim.height + ")");
			java.awt.image.PixelGrabber pg = new java.awt.image.PixelGrabber(imageStore, 0, 0, dim.width, dim.height, false);
			pg.grabPixels();
			Object objint = pg.getPixels();
			int[] pixels = (int[])objint;

			System.out.println("pixels.length = " + pixels.length);
//////////////////////
				FileOutputStream fos = new FileOutputStream("grafitoapp.dat");
				for(int i = 0; i < pixels.length; i++)
				{
					fos.write(pixels[i]);
				}
				fos.flush();
				fos.close();
//////////////////////

			GIFSave gifsave = new GIFSave(dim.width, dim.height, pixels);
			gifsave.writeImageToFile(new FileOutputStream(filename));
		} 
		catch(FileNotFoundException fnfe)
		{
			System.out.println(fnfe);
		}
		catch(IOException ioe)
		{
			System.out.println(ioe);
		}
		catch(Exception e)
		{
			System.out.println(e);
		}
	}

	private void loadParams(String paramFile)
	{
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(paramFile));
			String lineIn = br.readLine();
			while(lineIn != null)
			{
				if(lineIn.indexOf(":") > -1)
				{
					String key = lineIn.substring(0, lineIn.indexOf(":"));
					String val = lineIn.substring(lineIn.indexOf(":") + 1, lineIn.length());
					htParams.put(key, val);
				}
				lineIn  = br.readLine();
			}
		}
		catch(Exception e)
		{
			e.printStackTrace(System.out);
		}
	}

	private String getParameter(String paramName)
	{
		if(htParams.containsKey(paramName))
		{
			return (String)(htParams.get(paramName));
		}
		return (String)null;
	}

	/* Parse a "rXXXgXXXbXXX" String into a Color */
	private Color parseColor(String sColor)
	{
		int rVal = 0;
		int gVal = 0;
		int bVal = 0;
		try
		{
			rVal = Integer.parseInt(sColor.substring(sColor.indexOf("r") + 1, sColor.indexOf("g")));
			gVal = Integer.parseInt(sColor.substring(sColor.indexOf("g") + 1, sColor.indexOf("b")));
			bVal = Integer.parseInt(sColor.substring(sColor.indexOf("b") + 1, sColor.length()));
		}
		catch(NumberFormatException nfe)
		{
			return Color.gray;
		}
		return new Color(rVal, gVal, bVal);
	}

	/* Load an image */
	private Image loadImage(String imgname)
	{
		Image imgTemp = Toolkit.getDefaultToolkit().getImage(imgname);
		mt.addImage(imgTemp, 1);
		try
		{
			mt.waitForID(1);
		}
		catch(InterruptedException ie)
		{
			ie.printStackTrace(System.out);
		}
		return imgTemp;
	}

	/* Overloaded paint method */
	public void paint(Graphics g)
	{
		super.paint(g);
		if(imageStore != null)
		{
			Graphics gfx = canvas.getGraphics();
			gfx.drawImage(imageStore, 0, 0, this);
		}
	}

	/* Main method */
	public static void main(String[] args)
	{
		if(args.length < 1)
		{
			Grafito grafito = new Grafito(null);
		}
		else
		{
			Grafito grafito = new Grafito(args[0]);
		}
	}

	/* Inner Class */
	class GrafitoCanvas extends Panel
	{
		private Color localColor;
		private Image localImage;

		public GrafitoCanvas(Color c, Image i)
		{
			super();
			localColor = c;
			localImage = i;
		}

		public GrafitoCanvas()
		{
			this(Color.white, (Image)null);
		}

		public void updateColor(Color c)
		{
			localColor = c;
		}

		public void updateImage(Image i)
		{
			localImage = i;
			paint(this.getGraphics());
		}

		public void clearCanvas()
		{
			Graphics g = this.getGraphics();
			g.setColor(localColor);
			g.fillRect(0, 0, Grafito.this.cnvsXDimension, Grafito.this.appYDimension);
			paint(this.getGraphics());
		}

		public void paint(Graphics g)
		{
			g = this.getGraphics();
			if(localImage != null)
			{
				g.drawImage(localImage, 0, 0, this);
			}
		}
	}

}