/*
GNU General Public License

GrafitoApplet - Simple Grafitti Board Applet
Copyright (C) 2000/2001  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.grafito;

import java.applet.Applet;
import java.awt.Button;
import java.awt.Checkbox;
import java.awt.CheckboxGroup;
import java.awt.Choice;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.Label;
import java.awt.MediaTracker;
import java.awt.Panel;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.PixelGrabber;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.net.URL;
import java.net.URLConnection;

import com.hexidec.grafito.Constants;

/** GrafitoApplet
  * Applet for providing a simple grafitti board
  *
  * @author Howard Kistler
  * @version 0.3
  *
  * REQUIREMENTS
  * JDK 1.3 or higher
  *
  * VERSION HISTORY
  * 0.1  (09/14/2000) - initial creation (09/14/2000)
  * 0.2  (10/20/2000) - added GrafitoCanvas to better control painting
  *                   - added background image chooser
  *                   - implemented MediaTracker to make sure images preload (more important than it seems)
  * 0.3  (02/06/2001) - works with ImageServlet to write out a GIF on the server
  */

public class GrafitoApplet extends Applet implements ActionListener, MouseListener, MouseMotionListener, ItemListener
{
	// Drawing canvas
	private GrafitoCanvas canvas = new GrafitoCanvas();

	// Image to store drawing
	private Image imageStore;

	// Optional background image
	private Image imageBack;
	private int currImage = 0;

	// Default app size variables
	protected int appXDimension  = 500;
	protected int appYDimension  = 400;
	protected int menuXDimension = 100;
	protected int cnvsXDimension = appXDimension - menuXDimension;

	// Current paint setting variables
	private Color currColor = Color.black;
	private Color backColor = Color.white;

	// UI components
	private Checkbox chkBrushSmall;
	private Checkbox chkBrushMedium;
	private Checkbox chkBrushLarge;
	private Choice   chsPaintColor;
	private Choice   chsBackground;

	// Collections
	private Color[] forecolors;
	private Image[] backgrounds;

	/* Constructor */
	public void init()
	{
		int canvasWidth  = -1;

		// Fetch parameters
		String paramAppwidth     = getParameter("APPWIDTH");
		String paramAppheight    = getParameter("APPHEIGHT");
		String paramMenuwidth    = getParameter("MENUWIDTH");
		String paramCanvaswidth  = getParameter("CANVASWIDTH");
		String paramImagenames   = getParameter("IMAGENAMES");
		String paramImagefiles   = getParameter("IMAGEFILES");
		String paramImageback    = getParameter("IMAGEBACK");
		String paramColorback    = getParameter("COLORBACK");
		String paramColornames   = getParameter("COLORNAMES");
		String paramColorvalues  = getParameter("COLORVALUES");

		// Process parameters
		if(paramAppwidth != null)
		{
			try
			{
				appXDimension = Integer.parseInt(paramAppwidth);
				cnvsXDimension = appXDimension - menuXDimension;
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(paramAppheight != null)
		{
			try
			{
				appYDimension = Integer.parseInt(paramAppheight);
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(paramMenuwidth != null)
		{
			try
			{
				menuXDimension = Integer.parseInt(paramMenuwidth);
				cnvsXDimension = appXDimension - menuXDimension;
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(paramCanvaswidth != null)
		{
			try
			{
				canvasWidth = Integer.parseInt(paramCanvaswidth);
			}
			catch(NumberFormatException nfe) { ; }
		}
		if(canvasWidth > -1)
		{
			cnvsXDimension = canvasWidth;
			appXDimension  = cnvsXDimension + menuXDimension;
		}
		if((paramImagenames != null) && (paramImagefiles != null))
		{
			MediaTracker mt = new MediaTracker(this);
			chsBackground = new Choice();
			chsBackground.addItemListener(this);
			chsBackground.add("None");

			// Parse image names
			while(paramImagenames.indexOf("|") > -1)
			{
				String iName = paramImagenames.substring(0, paramImagenames.indexOf("|"));
				chsBackground.add(iName);
				paramImagenames = paramImagenames.substring(paramImagenames.indexOf("|") + 1, paramImagenames.length());
			}
			if(paramImagenames.length() > 0)
			{
				chsBackground.add(paramImagenames);
			}

			// Load image files
			backgrounds = new Image[chsBackground.getItemCount()];
			int imagecount = 0;
			backgrounds[imagecount] = (Image)null;
			imagecount++;
			while((paramImagefiles.indexOf("|") > -1) && (imagecount < chsBackground.getItemCount()))
			{
				String iFile = paramImagefiles.substring(0, paramImagefiles.indexOf("|"));
				Image tempImage = getImage(getDocumentBase(), iFile);
				mt.addImage(tempImage, 0);
				try
				{
					mt.waitForID(0);
				}
				catch(Exception e)
				{
				}
				backgrounds[imagecount] = tempImage;
				imagecount++;
				paramImagefiles = paramImagefiles.substring(paramImagefiles.indexOf("|") + 1, paramImagefiles.length());
			}
			if((paramImagefiles.length() > 0) && (imagecount < chsBackground.getItemCount()))
			{
				Image tempImage = getImage(getDocumentBase(), paramImagefiles);
				mt.addImage(tempImage, 0);
				try
				{
					mt.waitForID(0);
				}
				catch(Exception e)
				{
				}
				backgrounds[imagecount] = tempImage;
				imagecount++;
			}
		}
		else if(paramImageback != null)
		{
			MediaTracker mt = new MediaTracker(this);
			Image tempImage = getImage(getDocumentBase(), paramImageback);
			mt.addImage(tempImage, 0);
			try
			{
				mt.waitForID(0);
			}
			catch(Exception e)
			{
			}
			imageBack = tempImage;
		}
		if(paramColorback != null)
		{
			backColor = new Color(Integer.parseInt(paramColorback,16));
		}

		/* Set up the controls */
		Button btnFileRefresh = new Button("Refresh"); btnFileRefresh.setActionCommand("refresh"); btnFileRefresh.addActionListener(this);
		Button btnFileClear   = new Button("Clear");   btnFileClear.setActionCommand("clear");     btnFileClear.addActionListener(this);
		Button btnImageSave   = new Button("Save");    btnImageSave.setActionCommand("save");      btnImageSave.addActionListener(this);

		Label lblBrush = new Label("BRUSH", Label.CENTER);

		CheckboxGroup cgrpBrushes = new CheckboxGroup();
		chkBrushSmall  = new Checkbox("Small",  cgrpBrushes, false);
		chkBrushMedium = new Checkbox("Medium", cgrpBrushes, true);
		chkBrushLarge  = new Checkbox("Large", cgrpBrushes, false);

		Label lblColor = new Label("COLOR", Label.CENTER);

		if((paramColornames != null) && (paramColorvalues != null))
		{
			chsPaintColor = new Choice();
			chsPaintColor.addItemListener(this);

			// Parse color names
			while(paramColornames.indexOf("|") > -1)
			{
				String cName = paramColornames.substring(0, paramColornames.indexOf("|"));
				chsPaintColor.add(cName);
				paramColornames = paramColornames.substring(paramColornames.indexOf("|") + 1, paramColornames.length());
			}
			if(paramColornames.length() > 0)
			{
				chsPaintColor.add(paramColornames);
			}

			// Parse color values
			forecolors = new Color[chsPaintColor.getItemCount()];
			int colorcount = 0;
			while((paramColorvalues.indexOf("|") > -1) && (colorcount < chsPaintColor.getItemCount()))
			{
				String cValue = paramColorvalues.substring(0, paramColorvalues.indexOf("|"));
				forecolors[colorcount] = new Color(Integer.parseInt(cValue,16));
				colorcount++;
				paramColorvalues = paramColorvalues.substring(paramColorvalues.indexOf("|") + 1, paramColorvalues.length());
			}
			if((paramColorvalues.length() > 0) && (colorcount < chsPaintColor.getItemCount()))
			{
				forecolors[colorcount] = new Color(Integer.parseInt(paramColorvalues,16));
				colorcount++;
			}

			chsPaintColor.select(0);
			currColor = forecolors[0];
		}

		Panel pnlButtons = new Panel();
		pnlButtons.setBackground(new Color(232, 232, 232));
		int rowcount = 7 + (chsPaintColor == null ? 0 : 2) + (chsBackground == null ? 0 : 2);
		pnlButtons.setLayout(new GridLayout(rowcount, 1));
		if(chsBackground != null)
		{
			Label lblBack = new Label("IMAGE", Label.CENTER);
			pnlButtons.add(lblBack);
			pnlButtons.add(chsBackground);
		}
		pnlButtons.add(lblBrush);
		pnlButtons.add(chkBrushSmall);
		pnlButtons.add(chkBrushMedium);
		pnlButtons.add(chkBrushLarge);
		if(chsPaintColor != null)
		{
			pnlButtons.add(lblColor);
			pnlButtons.add(chsPaintColor);
		}
		pnlButtons.add(btnFileClear);
		pnlButtons.add(btnFileRefresh);
		pnlButtons.add(btnImageSave);
		pnlButtons.setSize(menuXDimension, appYDimension);
		pnlButtons.setBounds(0, 0, menuXDimension, appYDimension);

		canvas.addMouseListener(this);
		canvas.addMouseMotionListener(this);
		canvas.setSize(cnvsXDimension, appYDimension);
		canvas.setBounds(menuXDimension, 0, cnvsXDimension, appYDimension);
		canvas.setBackground(backColor);
		canvas.updateColor(backColor);
		canvas.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));

		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.setLayout(null);
		this.add(canvas);
		this.add(pnlButtons);
	}

	public void start()
	{
		initForPaint();
	}

	/* ActionListener method */
	public void actionPerformed(ActionEvent ae)
	{
		if(ae.getActionCommand().equals("clear"))
		{
			clearImage();
		}
		else if(ae.getActionCommand().equals("refresh"))
		{
			repaint();
		}
		else if(ae.getActionCommand().equals("save"))
		{
			serToServer();
		}
	}

	/* MouseListener methods */
	public void mousePressed(MouseEvent me)  { paintDot(me); }
	public void mouseClicked(MouseEvent me)  { ; }
	public void mouseEntered(MouseEvent me)  { ; }
	public void mouseExited(MouseEvent me)   { ; }
	public void mouseReleased(MouseEvent me) { ; }

	/* MouseMotionListener methods */
	public void mouseDragged(MouseEvent me) { paintDot(me); }
	public void mouseMoved(MouseEvent me)   { ; }

	/* ItemListener method */
	public void itemStateChanged(ItemEvent e)
	{
		if(e.getSource() == chsBackground)
		{
			int i = chsBackground.getSelectedIndex();
			if(i != currImage)
			{
				imageBack = backgrounds[i];
	
				clearImage();
			}
			currImage = i;
		}
		else if(e.getSource() == chsPaintColor)
		{
			currColor = forecolors[chsPaintColor.getSelectedIndex()];
		}
	}

	/* Initialise the image */
	public void initForPaint()
	{
		imageStore = (Image)createImage(cnvsXDimension, appYDimension); 
		clearImage();
	}

	/* Clear the image */
	public void clearImage()
	{
		Graphics gimg = imageStore.getGraphics();
		gimg.setColor(backColor);
		gimg.fillRect(0, 0, cnvsXDimension, appYDimension);
		if(imageBack != null)
		{
			gimg.drawImage(imageBack, 0, 0, this);
		}
		canvas.updateImage(imageStore);
		canvas.clearCanvas();
		paint(this.getGraphics());
	}

	/* Draw on the canvas */
	private void paintDot(MouseEvent me)
	{
		if(imageStore == null) { initForPaint(); }
		Graphics gimg = imageStore.getGraphics();
		gimg.setColor(currColor);

		if(chkBrushSmall.getState())
		{
			int minX = me.getX();
			int minY = me.getY();
			if(minX < 0) { minX = 0; }
			if(minY < 0) { minY = 0; }
			if(minX > cnvsXDimension) { minX = cnvsXDimension; }
			if(minY > appYDimension)  { minY = appYDimension; }
			gimg.drawLine(me.getX(), me.getY(), me.getX(), me.getY());
		}
		else if(chkBrushMedium.getState())
		{
			int minX = me.getX() - 2;
			int minY = me.getY() - 2;
			int maxX = 5;
			int maxY = 5;
			if(minX < 0) { minX = 0; }
			if(minY < 0) { minY = 0; }
			if(me.getX() + maxX > cnvsXDimension) { maxX = cnvsXDimension; }
			if(me.getY() + maxY > appYDimension)  { maxY = appYDimension; }
			gimg.fillRect(minX, minY, maxX, maxY);
		}
		else if(chkBrushLarge.getState())
		{
			int minX = me.getX() - 4;
			int minY = me.getY() - 4;
			int maxX = 9;
			int maxY = 9;
			if(minX < 0) { minX = 0; }
			if(minY < 0) { minY = 0; }
			if(me.getX() + maxX > cnvsXDimension) { maxX = cnvsXDimension; }
			if(me.getY() + maxY > appYDimension)  { maxY = appYDimension; }
			gimg.fillRect(minX, minY, maxX, maxY);
		}
		canvas.updateImage(imageStore);
	}

	/* Overloaded paint method */
	public void paint(Graphics g)
	{
		super.paint(g);
		if(imageStore == null)
		{
			initForPaint();
		}
		if(imageStore != null)
		{
			canvas.updateImage(imageStore);
		}
	}

///////////////////////
	public void serToServer()
	{
		paint(this.getGraphics());
		try
		{
			URL servlet = new URL(getCodeBase(), Constants.IMAGE_SERVLET);
			URLConnection conn = servlet.openConnection();
			conn.setDoInput(true);
			conn.setDoOutput(true);
			conn.setUseCaches(false);

			ObjectOutputStream out = new ObjectOutputStream(conn.getOutputStream());
			Dimension dimPic = new Dimension(imageStore.getWidth(this), imageStore.getHeight(this));
			out.writeObject(dimPic);

			int[] PixelMap = new int[dimPic.width * dimPic.height];
			PixelGrabber pg = new PixelGrabber(imageStore, 0, 0, dimPic.width, dimPic.height, PixelMap, 0, dimPic.width);
			pg.grabPixels();
			out.writeObject(PixelMap);

			out.flush();
			out.close();

			InputStream is = conn.getInputStream();
		}
		catch(Exception e)
		{
			e.printStackTrace(System.out);
		}
	}
///////////////////////

	/* Inner Class */
	class GrafitoCanvas extends Panel
	{
		private Color localColor;
		private Image localImage;

		public GrafitoCanvas(Color c, Image i)
		{
			super();
			localColor = c;
			localImage = i;
		}

		public GrafitoCanvas()
		{
			this(Color.white, (Image)null);
		}

		public void updateColor(Color c)
		{
			localColor = c;
		}

		public void updateImage(Image i)
		{
			localImage = i;
			paint(this.getGraphics());
		}

		public void clearCanvas()
		{
			Graphics g = this.getGraphics();
			g.setColor(localColor);
			g.fillRect(0, 0, GrafitoApplet.this.cnvsXDimension, GrafitoApplet.this.appYDimension);
			paint(this.getGraphics());
		}

		public void paint(Graphics g)
		{
			g = this.getGraphics();
			if(localImage != null)
			{
				g.drawImage(localImage, 0, 0, this);
			}
		}
	}

}